from livereload import Server
import os
import shutil
import argparse


class Kart:
    def __init__(self, miners=[], mappers=[], modifiers=[], renderers=[], config={}):
        self.miners = miners
        self.mappers = mappers
        self.modifiers = modifiers
        self.renderers = renderers
        self.config = config

    def prepare(self):
        self.site = {}
        for miner in self.miners:
            self.site.update(miner.collect())
        self.site["config"] = self.config
        for modifier in self.modifiers:
            modifier.modify(self.site)

    def create_map(self):
        self.map = {}
        for mapper in self.mappers:
            self.map.update(mapper.map(self.site))

    def write(self):
        for renderer in self.renderers:
            renderer.url_function = self.url
            renderer.render(self.map, self.site, self.build_location)

    def url(self, *name):
        name = ".".join(name)
        if not name:
            return ""
        if name in self.map.keys():
            result = self.map[name]["url"]
        elif name + ".1" in self.map.keys():
            result = self.map[name + ".1"]["url"]
        elif "/" in name:
            result = name
        else:
            return ""
        return self.config["base_url"] + result

    def build(self, build_location="_site"):
        self.build_location = build_location
        self.prepare()
        self.create_map()
        shutil.rmtree(self.build_location, ignore_errors=True)
        if "root" in os.listdir():
            shutil.copytree("root", self.build_location)
        else:
            os.makedirs(self.build_location, exist_ok=True)
        if "static" in os.listdir():
            shutil.copytree("static", os.path.join(self.build_location, "static"))
        self.write()

    def serve(self, port=9000):
        self.build()
        server = Server()
        server.watcher.ignore_dirs("_site")
        server.watch(".", self.build)
        server.serve(root=self.build_location, port=port, host="0.0.0.0")

    def run(self):
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "command", help="command to execute", choices={"build", "serve"}
        )
        parser.add_argument(
            "-p", "--port", help="port to bind to", default=9000, type=int
        )
        parser.add_argument("--production", action="store_true")
        args = parser.parse_args()
        if not args.production:
            self.config["base_url"] = ""
        if args.command == "build":
            self.build()
        if args.command == "serve":
            self.build_location = "_site"
            self.serve(args.port)



from bs4 import BeautifulSoup


class RuleModifier:
    def __init__(self, rules=[]):
        self.rules = rules  # a rule is a function

    def modify(self, site):
        for rule in self.rules:
            rule(site)


class TagModifier:
    def modify(self, site):
        site["tag_dict"] = {}
        for tag in site["tags"]:
            site["tag_dict"][tag["slug"]] = tag["name"]


class CollectionSorter:
    def __init__(self, collection_name, key, reverse=False):
        self.collection_name = collection_name
        self.key = key
        self.reverse = reverse

    def modify(self, site):
        site[self.collection_name].sort(key=lambda x: x[self.key])
        if self.reverse:
            site[self.collection_name].reverse()


class TocModifier:
    def __init__(self, collection_name):
        self.collection_name = collection_name

    def modify(self, site):
        if self.collection_name == "pages":
            items = list(site["pages"].values())
        else:
            items = site[self.collection_name]
        for item in items:
            toc = []
            soup = BeautifulSoup(item["content"], features="lxml")
            current_list = toc
            previous_tag = None
            for header in soup.findAll(["h2", "h3"]):
                if previous_tag == "h2" and header.name == "h3":
                    current_list = []
                elif previous_tag == "h3" and header.name == "h2":
                    toc.append(current_list)
                    current_list = toc
                current_list.append((header["id"], header.string))
                previous_tag = header.name
            if current_list != toc:
                toc.append(current_list)
            item["toc"] = toc



from jinja2 import Environment, FileSystemLoader
from feedgen.feed import FeedGenerator
import os
from datetime import timezone, time, datetime
import xml.etree.ElementTree as xml


class DefaultSiteRenderer:
    def __init__(
        self,
        url_function=None,
        name="default_site_renderer",
        template_folder="templates",
    ):
        self.name = name
        self.template_folder = template_folder
        self.url_function = url_function
        self.env = Environment(loader=FileSystemLoader(self.template_folder))

    def date_to_string(self, date):
        return date.strftime("%b %d, %Y")

    def render_single(self, page, site):
        if self.name != page["renderer"]:
            return
        template = ""
        if "template" in page["data"].keys():
            template = page["data"]["template"]
        if not template:
            template = page["default_template"]
        jinja_template = self.env.get_template(template)
        page["data"]["url"] = page["url"]

        jinja_template.globals.update(
            url=self.url_function, date_to_string=self.date_to_string
        )
        return jinja_template.render(page=page["data"], site=site)

    def render(self, map, site, build_location="_site"):
        self.map = map
        for page in map.values():
            rendered_file = self.render_single(page, site)
            if rendered_file:
                os.makedirs(build_location + page["url"], exist_ok=True)
                with open(build_location + page["url"] + "index.html", "w") as f:
                    f.write(rendered_file)


class DefaultFeedRenderer:
    def __init__(self, name="default_feed_renderer", collections=["posts"]):
        self.name = name
        self.collections = collections
        self.url_function = None

    def render_single(self, page, site):
        return

    def render(self, map, site, build_location="_site"):
        base_url = site["config"]["base_url"]
        fg = FeedGenerator()
        fg.title(site["config"]["name"])
        if not base_url:
            fg.id("base_url/")
        else:
            fg.id(base_url)
        fg.link({"href": base_url})
        fg.link({"href": base_url + "/atom.xml", "rel": "self"})

        feed_entries = []

        for collection in self.collections:
            for object in site[collection]:
                feed_entries.append([collection, object])

        feed_entries.sort(key=lambda x: x[1]["date"])

        for collection, entry in feed_entries:
            fe = fg.add_entry()
            if "title" in entry.keys():
                fe.title(entry["title"])
            elif "name" in entry.keys():
                fe.title(entry["name"])
            if "description" in entry.keys():
                fe.description(entry["description"])
            fe.updated(datetime.combine(entry["date"], time(12), tzinfo=timezone.utc))
            fe.id(self.url_function(collection, entry["slug"]))
            fe.link({"href": self.url_function(collection, entry["slug"])})
        fg.atom_file(build_location + "/atom.xml")


class DefaultSitemapRenderer:
    def __init__(self, name="default_sitemap_renderer"):
        self.name = name

    def render(self, map, site, build_location="_site"):
        base_url = site["config"]["base_url"]
        root = xml.Element("urlset")
        root.set("xmlns", "http://www.sitemaps.org/schemas/sitemap/0.9")
        for x in [
            x["url"] for x in map.values() if x["renderer"] == "default_site_renderer"
        ]:
            url = xml.SubElement(root, "url")
            loc = xml.SubElement(url, "loc")
            loc.text = base_url + x
        with open(build_location + "/sitemap.xml", "w") as f:
            f.write(
                '<?xml version="1.0" encoding="UTF-8"?>' + xml.tostring(root).decode()
            )
